#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <digit.h>

#define INCORRECT \
{\
	mValid = false;\
	return;\
}

#define NULL    0

/**************************************************************************************/
number::number( void )
:
	mValid( false )
{
}

/**************************************************************************************/
// Convert input string to double
number::number( char* inValue )
:
	mValid( false )
{
	if( !inValue )
		return;

	if( inValue[ 0 ] != 0 )
	{
		mValid = true;

		char* str = inValue;
		char* exp = NULL;
		char* base = NULL;

		exp = strchr( str, 'E' );
		if( exp )
		{
			if( exp == str )
				INCORRECT

			exp[ 0 ] = 0;
			++exp;

			if( !strlen( exp ) )
				INCORRECT

			if( exp[ 0 ] == 'B' )
				INCORRECT
		}

		char* bstart = exp ? exp : str;
		base = strchr( bstart, 'B' );
		if( base )
		{
			base[ 0 ] = 0;
			++base;

			if( !strlen( base ) )
				INCORRECT
		}

		errno = 0;
		mValue = atof( str );
		if( errno )
			INCORRECT

		// EXP
		if( exp )
		{
			if( strchr( exp, '.' ) )
				INCORRECT;

			errno = 0;
			long expValue = atoi( exp );
			if( errno )
				INCORRECT;

			if( expValue )
			{
				errno = 0;
				mValue *= pow( (double) 10, expValue );
				if( errno )
					INCORRECT;
			}
		}

		// Base
		if( base )
		{
			if( strchr( base, '.' ) )
				INCORRECT;

			errno = 0;
			int baseValue = atoi( base );
			if( errno )
				INCORRECT

			if( baseValue != 10 )
			{
				errno = 0;
				mValue = ChangeBase( mValue, baseValue );
				if( errno )
					INCORRECT
			}
		}
	}
}

/**************************************************************************************/
// Convert input string to double
double number::ChangeBase( double inValue, long inBase )
{
	double res = 0;

	if( inBase < 2 || inBase > 10 )
	{
		errno = 1;
		return res;
	}

	bool minus = ( inValue < 0 );
	inValue = fabs( inValue );

	long double idpart;
	long double fpart = modf( inValue, &idpart );
	long ipart = (long) idpart;
	long double mult = 1.;
	
	while( ipart > 0 )
	{
		long digit = ipart % 10;
		ipart /= 10;
		if( digit >= inBase )
		{
			errno = 1;
			return res;
		}

		res += mult * (double) digit;
		mult *= (double) inBase;
	}

	mult = 1. / double( inBase );

	while( fpart > 0 )
	{
		fpart *= (double) 10;
		long double ddigit;
		fpart = modf( fpart, &ddigit );
		long digit = (long) ddigit;

		res += mult * (double) digit;
		mult /= (double) inBase;
	}

	if( minus )
		res = -res;

	return res;
}

/**************************************************************************************/
void number::Print( void )
{
	if( mValid )
		printf( "%f\n", mValue );
	else
		printf( "NaN\n" );
}

/**************************************************************************************/
void number::Plus( const number& inValue )
{
	if( mValid )
	{
		if( inValue.GetIsValid() )
			mValue += inValue.GetDouble();
		else
			mValid = false;
	}
}

/**************************************************************************************/
void number::Minus( const number& inValue )
{
	if( mValid )
	{
		if( inValue.GetIsValid() )
			mValue -= inValue.GetDouble();
		else
			mValid = false;
	}
}

/**************************************************************************************/
void number::Mult( const number& inValue )
{
	if( mValid )
	{
		if( inValue.GetIsValid() )
			mValue *= inValue.GetDouble();
		else
			mValid = false;
	}
}

/**************************************************************************************/
void number::Div( const number& inValue )
{
	if( mValid )
	{
		if( inValue.GetIsValid() )
		{
			double value = inValue.GetDouble();
			if( value != 0.0 )
				mValue /= value;
			else
				mValid = false;
		}
		else
		{
			mValid = false;
		}
	}
}

/**************************************************************************************/
void number::Pow( const number& inValue )
{
	if( mValid )
	{
		if( inValue.GetIsValid() )
		{
			errno = 0;
			mValue = pow( mValue, inValue.GetDouble() );
			if( errno )
				mValid = false;
		}
		else
		{
			mValid = false;
		}
	}
}

/**************************************************************************************/
void number::Bigger( const number& inValue )
{
	if( mValid )
	{
		if( inValue.GetIsValid() )
			mValue = mValue > inValue.GetDouble() ? 1.0 : 0.0;
		else
			mValid = false;
	}
}

/**************************************************************************************/
void number::Smaller( const number& inValue )
{
	if( mValid )
	{
		if( inValue.GetIsValid() )
			mValue = mValue < inValue.GetDouble() ? 1.0 : 0.0;
		else
			mValid = false;
	}
}

/**************************************************************************************/
void number::Equal( const number& inValue )
{
	if( mValid )
	{
		if( inValue.GetIsValid() )
			mValue = mValue == inValue.GetDouble() ? 1.0 : 0.0;
		else
			mValid = false;
	}
}
