#include <calc.h>
#include <operator.h>

#include <parse.h>

/**************************************************************************************/
extern vector<Operator*>	operators;
extern vector<Operator*>	parents;
extern map<string,number>	variables;

long level = 0;

/**************************************************************************************/
Operator::Operator( 
	Operator* inParent, 
	const Type& inType, 
	const string& inExp,
	const string& inExp2 )
:
	mParent( inParent ),
	mCachedGoTo( NULL ),
	mType( inType ),
	mExp( inExp ),
	mExp2( inExp2 ),
	mCached( false )
{
}

/**************************************************************************************/
void Operator::AddChild( Operator* inOperator )
{
	if( inOperator )
		mChild.push_back( inOperator );
}

/**************************************************************************************/
bool Operator::Execute( void )
{
	bool res = true;

	++level;
	if( level > 512 )
		Exit( "Endless loop detected!" );

	switch( mType )
	{
		case IF:
		{
			size_t count = mChild.size();
			if( count == 1 )
			{
				if( !mCached )
					mCache = PreCalculate( mExp );

				number num = PostCalculate( mCache );
				if( num.GetIsValid() && num.GetDouble() != 0.0 )
					res = mChild[ 0 ]->Execute();
			}
		}
		break;

		case WHILE:
		{
			size_t count = mChild.size();
			if( count == 1 )
			{
				Operator* child = mChild[ 0 ];

				if( !mCached )
					mCache = PreCalculate( mExp );

				number num = PostCalculate( mCache );
				while( num.GetIsValid() && num.GetDouble() != 0.0 )
				{
					if( !mChild[ 0 ]->Execute() )
					{
						res = false;
						break;
					}

					num = PostCalculate( mCache );
				}
			}
		}
		break;

		case BLOCK:
		{
			size_t count = mChild.size();
			for( size_t i = 0 ; i < count ; ++i )
			{
				if( !mChild[ i ]->Execute() )
				{
					res = false;
					break;
				}
			}
		}
		break;

		case PRINT:
		{
			if( !mCached )
				mCache = PreCalculate( mExp );

			PostCalculate( mCache ).Print();
		}
		break;

		case LET:
		{
			if( !mCached )
				mCache = PreCalculate( mExp2 );

			variables[ mExp ] = PostCalculate( mCache );
		}
		break;

		case GOTO:
		{
			if( !mCachedGoTo )
			{
				size_t count = operators.size();
				for( size_t i = 0 ; i < count ; ++i )
				{
					Operator* o = operators[ i ];
					if( o && o->GetKind() == LABEL && o->GetData() == mExp )
					{
						mCachedGoTo = o;
						break;
					}
				}
			}

			if( !mCachedGoTo )
				Exit( "Goto without label!" );

			Operator* parent = mCachedGoTo->GetParent();
			if( !parent )
			{
				Exit( "Runtime error - label without parent!" );
			}
			else
			{
				parent->Catch( mCachedGoTo );
				res = false;
			}
		}
		break;
	}

	--level;

	return res;
}

/**************************************************************************************/
void Operator::Catch( Operator* inOperator )
{
	bool start = false;

	size_t count = mChild.size();
	for( size_t i = 0 ; i < count ; ++i )
	{
		Operator* child = mChild[ i ];
		if( child == inOperator || start )
		{
			mChild[ i ]->Execute();
			start = true;
		}
	}
}
