#include <calc.h>
#include <digit.h>
#include <parse.h>

#define NULL		0

#define POST		if( post.size() )\
					{\
						string copy = post;\
						post = "";\
						ProcessToken( copy, text );\
					}

long				line = 1;
long				first = 0;
Operator*			father = NULL;
vector<Operator*>	operators;
vector<Operator*>	parents;
map<string,number>	variables;
string				cur;
string				post;

/**************************************************************************************/
void AddOperator( const Type& inType, const string& inExp, const string& inExp2 )
{
	Operator* parent = NULL;

	if( father )
	{
		parent = father;
		father = NULL;
	}
	else
	{
		if( !parents.empty() )
			parent = parents[ parents.size() - 1 ];
	}
	
	if( !parent )
	{
		++first;
		if( first > 1 )
			Exit( "Must be only one first level operator!" );
	}

	Operator* o = new Operator( parent, inType, inExp, inExp2 );

	operators.push_back( o );

	if( parent )
		parent->AddChild( o );

	switch( inType )
	{
		case BLOCK:
		{
			parents.push_back( o );
		}
		break;

		case IF:
		case WHILE:
		{
			father = o;
		}
		break;

		default:
			break;
	}
}

/**************************************************************************************/
void Exit( char* msg )
{
	if( line )
	{
		printf( "line: %d \n", line );
		printf( cur.c_str() );
		printf( "\n" );
	}

	printf( msg );

	getc( stdin );
	exit( 0 );
}

/**************************************************************************************/
string PopArgument( char* &text, bool operatorNext )
{
	string res;

	string left = NextToken( text );
	if( left != "(")
		Exit( "'(' - Not found!" );

	long level = 1;

	string tkn = NextToken( text );
	while( tkn != ")" || level != 1 )
	{
		if( tkn == "(" )
			++level;

		if( tkn == ")" )
			--level;

		if( !tkn.length() )
			Exit( "')' - Not found!" );

		res += tkn;

		tkn = NextToken( text );
	}

	if( level != 1 )
		Exit( "Braces mismatched!" );

	if( !res.length() )
		Exit( "Argument not found!" );

	left = NextToken( text );
	if( !left.length() )
		Exit( "Unexpected end of file!" );

	if( left == ";" && operatorNext )
		Exit( "Found ';' - but operator expected!" );
	else if( left.length() && left != ";" && left != "}" && !operatorNext )
		Exit( "';' missed!" );

	if( left.length() )
		post = left;

	return res;
}

/**************************************************************************************/
string PopExpression( char* &text )
{
	string res;

	string tkn = NextToken( text );
	while( tkn.length() && tkn != ";" )
	{
		if( tkn == "{" || tkn == "=" )
			Exit( "';' missed!" );
		
		if( tkn == "}" )
		{
			post = tkn;
			break;
		}

		res += tkn;

		tkn = NextToken( text );
	}

	if( tkn == ";" && !res.length() )
		Exit( "Expected expression, but ';' found!" );

	if( !res.length() )
		Exit( "Unexpected end of file!" );

	return res;
}

/**************************************************************************************/
void CheckName( const string& name )
{
	size_t len = name.length();
	if( !len )
		Exit( "Identifier expected!" );

	long opened = 0;
	string sub;

	for( size_t i = 0 ; i < len ; ++i )
	{
		char ch = name[ i ];
		if( ch >= 'A' && ch <= 'Z' ||
			ch >= 'a' && ch <= 'z' ||
			ch == '_' )
		{
			sub += ch;
		}
		else
		{
			if( i && ch == '[' || ch == ']' )
			{
				if( ch == '[' )
				{
					sub = "";
					++opened;
				}

				if( ch == ']' )
				{
					if( name[ i - 1 ] == '[' )
						Exit( "Empty braces!" );

					if( variables.find( sub ) == variables.end() )
						Exit( "Variable for [] must be defined first!" );

					--opened;
				}

				if( opened < 0 )
					Exit( "Braces mismatch!" );
			}
			else
			{
				if( i == 0 || ch < '0' || ch > '9' )
					Exit( "Syntax error!" );

				sub += ch;
			}
		}
	}

	if( opened != 0 )
		Exit( "Braces mismatch!" );
}

/**************************************************************************************/
void Print( char* &text )
{
	AddOperator( PRINT, PopArgument( text, false ) );
	POST
}

/**************************************************************************************/
void If( char* &text )
{
	AddOperator( IF, PopArgument( text, true ) );
	POST
}

/**************************************************************************************/
void While( char* &text )
{
	AddOperator( WHILE, PopArgument( text, true ) );
	POST
}

/**************************************************************************************/
void GoTo( char* &text )
{
	AddOperator( GOTO, PopArgument( text, false ) );
	POST
}

/**************************************************************************************/
void Block( char* &text )
{
	string empty;
	AddOperator( BLOCK, empty );
}

/**************************************************************************************/
void EndBlock( char* &text )
{
	if( parents.empty() )
	{
		Exit( "Syntax error - '}'" );
	}
	else
	{
		parents.pop_back();
	}
}

/**************************************************************************************/
void Let( const string& arg, const string& value )
{
	CheckName( arg );
	variables[ arg ] = number();
	AddOperator( LET, arg, value );
}

/**************************************************************************************/
void Include( char* &text )
{
	string res;

	string tkn = NextToken( text );
	while( tkn != "]" )
	{
		if( !tkn.length() )
			Exit( "']' - Not found!" );

		res += tkn;

		tkn = NextToken( text );
	}

	if( !res.length() )
		Exit( "Argument not found!" );

	AddOperator( INCLUDE, res );
}

/**************************************************************************************/
void Label( char* &text )
{
	string res;

	string tkn = NextToken( text );
	while( tkn.length() && tkn != ":" )
	{
		if( tkn == "{" || tkn == "}" || tkn == "=" )
			Exit( "':' missed!" );

		res += tkn;

		tkn = NextToken( text );
	}

	tkn = NextToken( text );
	if( tkn.length() && tkn != ";" )
		Exit( "';' expected!" );

	CheckName( res );

	AddOperator( LABEL, res );
}

/**************************************************************************************/
void ProcessToken( const string& inToken, char* &text )
{
	if( inToken == "print" )
	{
		Print( text );
	}
	else if( inToken == "goto" )
	{
		GoTo( text );
	}
	else if( inToken == "if" )
	{
		If( text );
	}
	else if( inToken == "while" )
	{
		While( text );
	}
	else if( inToken == "{" )
	{
		Block( text );
	}
	else if( inToken == "[" )
	{
		Include( text );
	}
	else if( inToken == "}" )
	{
		EndBlock( text );
	}
	else if( inToken == ":" )
	{
		Label( text );
	}
	else if( inToken == ";" )
	{
	}
	else
	{
		CheckName( inToken );

		string token = NextToken( text );
		if( token != "=" )
		{
			Exit( "'=' Expected" );
		}
		else
		{
			token = PopExpression( text );
			Let( inToken, token );
			POST
		}
	}
}

/**************************************************************************************/
string NextToken( char* &text )
{
	string res;
	
	while( text && text[ 0 ] != 0 )
	{
		char ch = text[ 0 ];

		// New line
		if( ch == '\n' || ch == '\r' )
		{
			cur = "";

			if( ch == '\n' )
				++line;

			++text;
			continue;
		}

		// Space
		if( ch == ' ' || ch == '\t' )
		{
			++text;
			cur += ch;
			continue;
		}

		// Quotes && =
		if( ch == '(' || ch == ')' || ch == '{' || ch == '}' || ch == '=' || ch == ':' )
		{
			if( res.length() )
			{
				break;
			}
			else
			{
				++text;
				res += ch;
				cur += ch;
				break;
			}
		}

		// End of token
		if( ch == ';' )
		{
			if( res.length() )
			{
				break;
			}
			else
			{
				++text;
				res += ch;
				cur += ch;
				break;
			}
		}
			
		// End of text
		if( text[ 0 ] == 0 )
			break;

		++text;
		cur += ch;
		res += ch;
	}

	return res;
}

/**************************************************************************************/
void Parse( char* text )
{
	line = 1;

	if( !text )
		Exit( "Empty file" );

	string token = NextToken( text );
	while( token.length() )
	{
		ProcessToken( token, text );
		token = NextToken( text );
	}

	if( parents.size() )
		Exit( "Unclosed '{'!" );

	line = 0;

	if( operators.size() )
		operators[ 0 ]->Execute();
	else
		Exit( "Must be at least one operator!" );

	getc( stdin );
}
